/*
** $Id: rthreads.c,v 4.4.3.1 1991/11/22 04:12:18 davison Trn $
**
** $Log: rthreads.c,v $
** Revision 4.4.3.1  1991/11/22  04:12:18  davison
** Trn Release 2.0
*/

/*
 * The trn database was pretty odd and embodied some very strange
 * and undocumented conventions.  These are recorded here for future
 * reference.  Much of the code here turns the newoverview database
 * into what the rest of trn expects to see.
 *
 * The arrays subject_ptrs and subject_cnts are parallel; they probably
 * ought to be a single array of a struct containing a pointer and a count.
 * More than one entry in subject_ptrs can point at the same subject.
 * The entries are grouped by root; there will be p_roots[0].subject_cnt
 * entries in both arrays, describing the subjects belonging to the first
 * root, followed by p_roots[1].subject_cnt entries for the next root, etc.
 * p_articles[].subject and root_subjects[] are indices into these arrays;
 * root_subjects is indexed by root number and points at the first entry
 * in each array for this root.
 */

#include "EXTERN.h"
#include "common.h"
#include "intrp.h"
#include "rn.h"
#include "ngdata.h"
#include "rcln.h"
#include "bits.h"
#include "util.h"
#include "rcstuff.h"
#ifdef SERVER
#include "server.h"
#endif

#undef USETMPTHREAD	/* force tobethreaded to be declared */

#include "threads.h"
#include "INTERN.h"
#include "rthreads.h"

#include <hash.h>
#include <hdbm.h>
#include <newsoverview.h>

#ifdef USETHREADS

/* imports */
extern char *strsave();

/* privates */
static int subjcnt, fromcnt, partidx, rootidx = 0;
static HASHTABLE *subjtbl;

/* forwards */
static int read_item();
static void safefree();

/*
 * stub for obsolete routine
 */
void
thread_init()
{
}

/*
 * Open a thread file to make use of the data it contains.  Everything is
 * slurped into arrays and buffers, and some minor digesting of the data
 * is performed to make it more palatable.  Be sure to call unuse_data()
 * before calling this again.
 */
int
use_data(thread_if_empty)
bool_int thread_if_empty;
{
	register int i;
	register struct novart *artp;
	ART_NUM last = getngsize(ng);
	ART_NUM first = getabsfirst(ng, last);
	struct novart *allovp;
	static struct novgroup *ngovp;

	if (last < first)
		return 0;			/* no articles */

	bzero((char *)&total, sizeof(TOTAL));
	if (ngovp != NULL)
		novclose(ngovp);		/* toss previous group */
	ngovp = novopen(ngname);
	if (ngovp == NULL) {
		printf("\n\nOpen failed for thread data.\n");
		return 0;			/* no articles overviewed */
	}
	allovp = novall(ngovp);
	if (allovp == NULL)
		return 0;
	novthread(ngovp);

	for (i = 0, artp = novall(ngovp); artp != NULL; artp = artp->a_nxtnum)
		i++;
	total.article = i;

	/* fill in total from allovp */
	total.first = first;
	total.last = last;
	total.string1 = 0;		/* obsolete */
	total.string2 = 0;		/* obsolete */
	total.root = total.article;	/* guess */
	total.subject = total.article;
	total.author = total.article;
	total.domain = total.article;

	/* opened overview data & read summary; set up */
	if (total.last < last && lastart > total.last) {
		tobethreaded = last - total.last;
		toread[ng] -= tobethreaded;
		lastart = total.last;
	}

	if (total.last > lastart) {
#ifdef SERVER				/* on an NNTP client, actually */
		if (time(Null(time_t * )) - lastactfetch > MINFETCHTIME) {
			fclose(actfp);
			ngdata_init();	/* re-grab the active file */
		}
#endif
		grow_ctl(total.last);	/* sets lastart */
		ngmax[ng] = lastart; /* ensure getngsize() knows the new max. */
	}

	/* allocate and fill in arrays of pointers to data just read */
#ifndef lint
	author_ptrs = (char **)safemalloc(total.article * sizeof(char **));
	author_cnts = (WORD *)safemalloc(total.article * sizeof(WORD));
	subject_ptrs = (char **)safemalloc(total.article * sizeof(char **));
	subject_cnts = (WORD *)safemalloc(total.article * sizeof(WORD));
	root_subjects = (WORD *)safemalloc(total.article * sizeof(WORD));
	root_article_cnts = (WORD *)safemalloc(total.article * sizeof(WORD));
	p_roots = (PACKED_ROOT *)
		safemalloc((MEM_SIZE)total.article * sizeof(PACKED_ROOT));
#endif
	p_articles = (PACKED_ARTICLE *)
		safemalloc((MEM_SIZE)total.article*sizeof(PACKED_ARTICLE));
	selected_roots = safemalloc(total.article * sizeof (char));

	bzero((char *)root_article_cnts, total.article * sizeof(WORD));
	bzero((char *)selected_roots, total.article * sizeof(char));
	bzero((char *)p_roots, total.article * sizeof(PACKED_ROOT));

	mklinks(ngovp);		/* fill in later article members */
	selectroots();

	/* Try to clean up the bitmap if articles are missing. */
	if (unthreaded)
		cleanbitmap();
	safefree(&saved_selections);
	select_page = 0;
	return 1;
}

/*
 * add subj to subject_ptrs & return index
 *
 * this should be as simple as fromidx, but, incredibly, trn actually
 * expects identical subjects to share an entry in subject_ptrs, and
 * relies upon this when drawing hunt-the-wumpus diagrams.
 */
static int
subjidx(subj)
char *subj;
{
	register int idx;
	char **subjent = (char **)hashfetch(subjtbl, subj);
	extern char subject_str[80];

	if (subjent != NULL) {
		idx = subjent - subject_ptrs;
		subject_cnts[idx]++;
	} else {
		idx = subjcnt++;
		subject_cnts[idx] = 1;
		get_subject_str(subj);	/* result in subject_str */
		subject_ptrs[idx] = strsave(subject_str);
		/* this better work */
		(void) hashstore(subjtbl, subj, (char *)&subject_ptrs[idx]);
	}
	return idx;
}

static int
fromidx(from)	/* add from to author_ptrs & return index */
char *from;
{
	register int idx = fromcnt++;
	extern char author_str[20];

	author_cnts[idx] = 1;
	get_author_str(from);		/* result in author_str */
	author_ptrs[idx] = strsave(author_str);
	return idx;
}

/*
 * It would be relatively straightforward to fill in a struct Article,
 * but noooo, we have to fill in a PACKED_ARTICLE.
 */
static
fillptart(ptartp, artp, flags)
register PACKED_ARTICLE *ptartp;
register struct novart *artp;
int flags;
{
	ptartp->num = atol(artp->a_num);
	ptartp->date = getdate(artp->a_date, NULL);
	ptartp->subject = subjidx(artp->a_subj); /* index of subject_ptrs */
	ptartp->author = fromidx(artp->a_from);	/* index of author_ptrs */
	ptartp->flags = flags;
	/* fill these in later */
	ptartp->child_cnt = 0;
	ptartp->parent = 0;		/* offset in p_articles to parent */
	ptartp->siblings = 0;		/* offset in p_articles to siblings */
	ptartp->root = 0;		/* index of p_roots */
}

static int
artidx(anumb)
char *anumb;
{
	register long numb = atol(anumb);
	register int i;

	for (i = 0; i < total.article; i++)
		if (p_articles[i].num == numb)
			return i;
	return -1;
}

static
fillptroot(ptrp, ptartp)
register PACKED_ROOT *ptrp;
register PACKED_ARTICLE *ptartp;
{
	ptrp->root_num = ptartp->num;
	ptrp->articles = ptartp - p_articles;
	/* correct these later */
	ptrp->thread_cnt = 0;
	ptrp->subject_cnt = 0;
}

static int ivisits = 0;		/* DEBUG */

static PACKED_ARTICLE *
insertart(gp, msgid, flags)
register struct novgroup *gp;
register char *msgid;
int flags;
{
	register struct novart *artp;
	register PACKED_ARTICLE *ptartp;

	if (msgid == NULL)
		return NULL;
	artp = (struct novart*)hashfetch(gp->g_msgids, msgid);
	if (artp == NULL) {
		errno = 0;
		warning("hashfetch failed in insertart() for `%s'", msgid);
		return NULL;
	}
	ptartp = &p_articles[partidx++];
	fillptart(ptartp, artp, flags);
	(void) insertart(gp, artp->a_child1, 0);
	(void) insertart(gp, artp->a_sibling, 0);
	return ptartp;
}

/*
 * set up a p_article entry for each root and its descendants,
 * and a p_roots entry for each root article
 */
/* ARGSUSED */
static
insertvisit(key, data, hook)
char *key, *data, *hook;
{
	register struct novart *artp = (struct novart *)data;
	register PACKED_ARTICLE *ptartp;

	ivisits++;	/* DEBUG */
	if (artp->a_msgid == NULL)
		printf("null message-id in insertvisit()\n");
	ptartp = insertart((struct novgroup *)hook, artp->a_msgid, ROOT_ARTICLE);
	if (ptartp != NULL)
		fillptroot(&p_roots[rootidx++], ptartp);
}

static struct novart *
msgtoart(gp, msgid)
struct novgroup *gp;
char *msgid;
{
	if (msgid == NULL)
		return NULL;
	return (struct novart *)hashfetch(gp->g_msgids, msgid);
}

static int avisits = 0;

/*
 * assuming p_roots has been initialised, set up parent and siblings
 * links and root members in p_articles.
 */
/* ARGSUSED */
static
artvisit(key, data, hook)
char *key, *data, *hook;
{
	register struct novart *oap;
	register struct novart *ap = (struct novart *)data;
	register int i;
	int idx = artidx(ap->a_num);
	register PACKED_ARTICLE *ptartp = &p_articles[idx];
	register char *msgid;
	register struct novgroup *gp = (struct novgroup *)hook;

	avisits++;
	if (idx < 0)
		error("can't find `%s'", ap->a_num);
	oap = msgtoart(gp, ap->a_parent);
	if (oap != NULL)
		ptartp->parent = artidx(oap->a_num) - idx;
	oap = msgtoart(gp, ap->a_sibling);
	if (oap != NULL)
		ptartp->siblings = artidx(oap->a_num) - idx;

	for (i = 0; i < total.root; i++)
		if (idx >= p_roots[i].articles && idx <
		    (i == total.root-1? total.article: p_roots[i+1].articles)) {
			ptartp->root = i;
			break;
		}

	ptartp->child_cnt = 0;
	for (msgid = ap->a_child1; msgid != NULL; msgid = oap->a_sibling) {
		oap = msgtoart(gp, msgid);
		if (oap == NULL)
			break;
		ptartp->child_cnt++;
	}
}

/*
 * descend trees, recompute .subject, root_subjects, subject_ptrs, subject_cnts
 */
static PACKED_ARTICLE *
subjart(gp, msgid, flags)
register struct novgroup *gp;
register char *msgid;
int flags;
{
	register struct novart *artp;
	register PACKED_ARTICLE *ptartp;
	int idx;

	if (msgid == NULL)
		return NULL;
	artp = (struct novart*)hashfetch(gp->g_msgids, msgid);
	if (artp == NULL) {
		errno = 0;
		warning("hashfetch failed in subjart() for `%s'", msgid);
		return NULL;
	}
	idx = artidx(artp->a_num);
	ptartp = &p_articles[idx];

	/* set subject_cnts, subject_ptrs & subjcnt too */
	ptartp->subject = subjidx(artp->a_subj); /* index of subject_ptrs */

	(void) subjart(gp, artp->a_child1, 0);
	(void) subjart(gp, artp->a_sibling, 0);
	return ptartp;
}

/*
 * visit roots, descend trees, recompute .subject, root_subjects,
 * subject_ptrs, subject_cnts, .subject_cnt
 */
/* ARGSUSED */
static
subjvisit(key, data, hook)
char *key, *data, *hook;
{
	register struct novart *artp = (struct novart *)data;
	int idx = artidx(artp->a_num);
	register PACKED_ARTICLE *ptartp = &p_articles[idx];
	int initcnt = subjcnt;

	if (artp->a_msgid == NULL)
		printf("null message-id in subjvisit()\n");
	if (subjtbl != NULL)
		hashdestroy(subjtbl);
	subjtbl = hashcreate(50, (unsigned (*)())NULL);
	if (subjtbl == NULL)
		return;
	(void) subjart((struct novgroup *)hook, artp->a_msgid, ROOT_ARTICLE);
	p_roots[ptartp->root].subject_cnt = subjcnt - initcnt;
}

/*
 * a subtle point: trn assumes that the p_articles array is filled in
 * tree-walk order, so that it can use its own peculiar numerology on the
 * indices, notably to decide where (sub)threads are.
 *
 * first hashwalk fills in p_articles and p_roots, fakes subject_cnt &
 * thread_cnt members, and sets up author_cnts, author_ptrs, subject_cnts,
 * and subject_ptrs.  the next hashwalk sets the parent, siblings, root
 * and child_cnt members in each p_articles entry, now that we can map
 * message-ids to article numbers and thence to p_articles indices.
 * the next hashwalk walks the trees from the roots and recomputes
 * subject and subject_cnt members and the contents of subject_ptrs,
 * and subject_cnts, now that root members are set.
 */
static
mklinks(ngovp)
register struct novgroup *ngovp;
{
	register int i;

	if (subjtbl != NULL)
		hashdestroy(subjtbl);
	subjtbl = hashcreate(50, (unsigned (*)())NULL);
	if (subjtbl == NULL)
		return;

	/* initialise p_articles and p_roots. */
	subjcnt = fromcnt = 0;
	rootidx = partidx = 0;
	hashwalk(ngovp->g_roots, insertvisit, (char *)ngovp);
	total.root = rootidx;

	/* correct thread_cnt members of roots */
	for (i = 0; i < total.root; i++) {
		register PACKED_ROOT *prp = &p_roots[i];

		prp->thread_cnt = (i == total.root-1? total.article:
			(prp+1)->articles) - prp->articles;	/* unused */
		prp->subject_cnt = 0;
	}

	/* correct parent, sibling, root and child_cnt members */
	hashwalk(ngovp->g_msgids, artvisit, (char *)ngovp);

	/*
	 * visit roots, descend trees, recompute .subject, root_subjects,
	 * subject_ptrs, subject_cnts, .subject_cnt
	 */
	subjcnt = 0;
	hashwalk(ngovp->g_roots, subjvisit, (char *)ngovp);
}

/*
 * walk through p_roots and set root_subjects and root_article_cnts
 * from subject_cnts.  If there are saved selections, set selected_roots
 * from p_roots, selected_root_cnt and saved_selections.
 */
static
selectroots()
{
	register int i, j, k;

	for (i = 0, j = 0; i < total.root; i++) {
		root_subjects[i] = j;
		for (k = p_roots[i].subject_cnt; k--; )
			root_article_cnts[i] += subject_cnts[j++];
		if (saved_selections != NULL)
			for (k = 0; k < selected_root_cnt; k++)
				if (p_roots[i].root_num==saved_selections[k]) {
					selected_roots[i] = 1;
					break;
				}
	}
	count_roots(saved_selections == NULL);
}

static
cleanbitmap()
{
	register long i;
	char *oldarea = ctlarea, *newarea;
	extern MEM_SIZE ctlsize;

	newarea = ctlarea = safemalloc(ctlsize);
	bzero((char *)ctlarea, ctlsize);
	for (i = total.article, p_art = p_articles; i--; p_art++)
		if (p_art->num >= firstbit)
			ctl_set(p_art->num);
	for (i = firstbit; i <= lastart; i++)
		if (!ctl_read(i)) {
			ctlarea = oldarea;
			oneless(i);
			ctlarea = newarea;
		}
	ctlarea = oldarea;
	free(newarea);
	p_art = Nullart;
	count_roots(FALSE);
}

/* Free some memory if it hasn't already been freed. */
static void
safefree(pp)
char **pp;
{
    if (*pp) {
	free(*pp);
	*pp = Nullch;
    }
}

/* Discard the thread data that we received through the use_data() call.
** If "save_selections" is non-zero, we'll try to remember which roots
** are currently selected long enough for the use_data() call to re-use
** them.  Only do this when you are going to re-open the same data file
** immediately with use_data() (presumably because the data has been
** updated while we were using it).
*/
void
unuse_data(save_selections)
bool_int save_selections;
{
	register int i, j;

	if (save_selections) {
#ifndef lint
		saved_selections = (ART_NUM *)
			safemalloc(selected_root_cnt * sizeof (ART_NUM));
#endif
		for (i = 0, j = 0; i < total.root; i++)
			if (selected_roots[i])
				saved_selections[j++] = p_roots[i].root_num;
	} else
		selected_root_cnt = selected_count = 0;
	safefree(&p_roots);
	safefree(&root_subjects);
	safefree(&author_cnts);
	safefree(&subject_cnts);
	for (i = 0; i < fromcnt; i++)
		free(author_ptrs[i]);
	safefree(&author_ptrs);
	for (i = 0; i < subjcnt; i++)
		free(subject_ptrs[i]);
	safefree(&subject_ptrs);
	safefree(&root_article_cnts);
	safefree(&selected_roots);
	safefree(&p_articles);

	p_art = curr_p_art = Nullart;
	init_tree();		/* free any tree lines */

	bzero((char *)&total, sizeof (TOTAL));
}

#endif /* USETHREADS */
